# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import shelve
import threading

import events
import errors
import item

from helper import debug

#
# ReadWriteLock from Python Cookbook Recipe 6.3
#

class ReadWriteLock:
    def __init__(self):
        self._read_ready = threading.Condition(threading.Lock())
        self._readers = 0

    def acquire_read(self):
        # debug("ReadWriteLock: acquire read\n")
        self._read_ready.acquire()
        try:
            self._readers += 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
        finally:
            self._read_ready.release()

    def release_read(self):
        # debug("ReadWriteLock: release read\n")
        self._read_ready.acquire()
        try:
            self._readers -= 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
            if not self._readers:
                self._read_ready.notifyAll()
        finally:
            self._read_ready.release()

    def acquire_write(self):
        # debug("ReadWriteLock: acquire write\n")
        self._read_ready.acquire()
        while self._readers>0:
            self._read_ready.wait()

    def release_write(self):
        # debug("ReadWriteLock: release write\n")
        self._read_ready.release()

class Lock:
    def __init__(self):
        self._lock = threading.Condition(threading.Lock())

    def acquire_read(self):
        import inspect
        debug("Lock: acquire read from %s\n" % inspect.stack()[1][3])
        self._lock.acquire()

    def release_read(self):
        import inspect
        debug("Lock: release read from %s\n" % inspect.stack()[1][3])
        self._lock.release()

    def acquire_write(self):
        import inspect
        debug("Lock: aquire write from %s\n" % inspect.stack()[1][3])
        self._lock.acquire()

    def release_write(self):
        import inspect
        debug("Lock: release write from %s\n" % inspect.stack()[1][3])
        self._lock.release()

#
# songdb class
#

def _albumid(artist, album):
    """ return unique id for given album of artist """
    return "|%s|%s|" % (artist, album)

class songdb:
    def __init__(self, channel, config):
        self.songdbbase = os.path.expanduser(config.songdbbase)
        self.basedir = os.path.expanduser(config.basedir)

        if not os.path.isdir(self.basedir):
            raise errors.configurationerror("config.basedir is not a directory.")

        if not os.access(self.basedir, os.X_OK | os.R_OK):
            raise errors.configurationerror("you are not allowed to access and read config.basedir.")
        
        if self.songdbbase:
            try:
                self.songs = shelve.open(self.songdbbase + "_songs.db")
                self.artists = shelve.open(self.songdbbase + "_artists.db")
                self.albums = shelve.open(self.songdbbase + "_albums.db")
                self.playlists = shelve.open(self.songdbbase + "_playlists.db")
                
                self.genres = shelve.open(self.songdbbase + "_genres.db")
                self.years = shelve.open(self.songdbbase + "_years.db")
            except:
                raise errors.databaseerror("cannot open song database files.")
            
            debug("songdb: %d songs, %d artists, %d albums, %d genres and %d playlists registered\n" %
                  (len(self.songs),
                   len(self.artists),
                   len(self.albums),
                   len(self.genres),
                   len(self.playlists)))
        else:
            self.songs = {}
            self.artists = {}
            self.albums = {}
            self.genres = {}
            self.years = {}
            self.playlists = {}

        # empty cache of all artists and all songs
        # This variables contain tuples acting as caches for
        # all artists and all songs, respectively, in the database
        self.allartists = self.allsongs = None

        # this lock protects write access to the database
        self.rwlock = ReadWriteLock()
        # self.rwlock = Lock()

        # we need to be informed about song info changes
        channel.subscribe(events.songupdate, self.songupdate)

        if config.autoregisterer:
            self.autoregisterer = songautoregisterer(self, self.basedir)
            self.autoregisterer.start()
        else:
            self.autoregisterer = None

    def __len__(self):
        self.rwlock.acquire_read()
        l = len(self.songs)
        self.rwlock.release_read()
        return l

    def close(self):
        if self.songdbbase:
            if self.autoregisterer:
                self.autoregisterer.stop = 1
                self.autoregisterer.join()

            self.rwlock.acquire_write()
            self.songs.close()
            self.artists.close()
            self.albums.close()
            self.genres.close()
            self.years.close()
            self.playlists.close()
            self.rwlock.release_write()

    # methods for insertion/update of respective index tables
    # Note that these methods assume that the caller already holds
    # a write lock on the database!

    def _reindexsong_album(self, song):
        """ insert/update album information for song """
        albumid = _albumid(song.artist, song.album)

        # marker: do we have to write new album information
        changed = 0

        # insert album in songdb, or fetch existent one from songdb
        if not self.albums.has_key(albumid):
            debug("songdb: new album: %s\n" % albumid)
            album = item.album(song.artist, song.album)
            changed = 1
            events.hub.notify(events.albumaddedtosongdb(album))
        else:
            album = self.albums[albumid]

        if song.path not in album.songs:
            changed = 1
            album.songs.append(song.path)
            
        if song.genre not in album.genres:
            changed = 1
            album.genres.append(song.genre)

        if song.year not in album.years:
            changed = 1
            album.years.append(song.year)

        if changed:
            self.albums[albumid] = album

    def _reindexsong_artist(self, song):
        """ insert/update artist information for song """
        albumid = _albumid(song.artist, song.album)
        changed = 0

        # insert artist in songdb, or fetch existent one from songdb
        if not self.artists.has_key(song.artist):
            debug("songdb: new artist: %s\n" % song.artist)
            artist = item.artist(song.artist)
            changed = 1
            events.hub.notify(events.artistaddedtosongdb(artist))
        else:
            artist = self.artists[song.artist]

        if albumid not in artist.albums:
            changed = 1
            artist.albums.append(albumid)

        if song.path not in artist.songs:
            changed = 1
            artist.songs.append(song.path)
        
        if song.genre not in artist.genres:
            changed = 1
            artist.genres.append(song.genre)

        if song.year not in artist.years:
            changed = 1
            artist.years.append(song.year)

        if changed:
            self.allartists = None
            self.artists[song.artist] = artist

    def _reindexsong_genre(self, song):
        """ insert/update genre information for song """
        
        changed = 0
                
        # insert genre in songdb, or fetch existent one from songdb
        if not self.genres.has_key(song.genre):
            debug("songdb: new genre: %s\n" % song.genre)
            genre = item.genre(song.genre)
            changed = 1
            # XXX no event
        else:
            genre = self.genres[song.genre]
            
        if song.artist not in genre.artists:
            changed = 1
            genre.artists.append(song.artist)

        if changed:
            self.genres[song.genre] = genre

    def _reindexsong_year(self, song):
        """ insert/update year information for song """

        yearid = str(song.year)
        changed = 0

        # insert year in songdb, or fetch existent one from songdb
        
        if not self.years.has_key(yearid):
            debug("songdb: new year: %s\n" % yearid)
            year = item.year(song.year)
            changed = 1
            # XXX no event
        else:
            year = self.years[yearid]

        if song.artist not in year.artists:
            changed = 1
            year.artists.append(song.artist)

        if changed:
            self.years[yearid] = year

    def queryregistersong(self, path):
        """get song info from database or insert new one"""

        self.rwlock.acquire_write()
        
        if self.songs.has_key(path):
            song = self.songs[path]
        else:
            debug("songdb: new song: %s\n" % path)
            song = self.songs[path] = item.song(path)
            self.allsongs = None

        # update indices
        self._reindexsong_album(song)
        self._reindexsong_artist(song)
        self._reindexsong_genre(song)
        self._reindexsong_year(song)

        self.rwlock.release_write()
            
        return song

    def delsong(self, path):
        """delete song from database"""

        # XXX: not very well tested

        self.rwlock.acquire_write()
        
        if not self.songs.has_key(path):
            raise "Cannot delete nonexistent song"

        debug("songdb: delete song: %s\n" % path)
        song = self.songs[path]

        # update artist and/or album info
        artist = self.artists[song.artist]
        album = artist.albums[song.album]
        
        artist.songs.remove(song.path)
        album.songs.remove(song.path)
        for asong in album.songs:
            if asong.genre==song.genre:
                break
        else:
            album.genres.remove(song.genre)
            for asong in artist.songs:
                if asong.genre==song.genre:
                    break
            else:
                artist.genres.remove(song.genre)
                
        self.artists[song.artist] = artist

        # delete album and artist, if necessary
        if len(album.songs)==0:
            del artist.albums[song.album]
            self.artists[song.artist] = artist

            if len(artist.albums)==0:
                del self.artists[song.artist]
                debug("songdb: deleted artist: %s\n" % song.artist)
                self.allartists = None
                # XXX: no event

        del self.songs[path]
        self.allsongs = None

        self.rwlock.release_write()

    def registerplaylist(self, path):
        playlist = item.playlist(path)

        # also try to register songs in playlist
        # delete song, if this fails
        for i in range(len(playlist.songs)):
            try:
                self.queryregistersong(playlist.songs[i])
            except (IOError, OSError):
                del playlist.songs[i]

        # a resulting, non-empty playlist can be written in the database
        if playlist.songs:
            self.rwlock.acquire_write()
            self.playlists[path] = playlist
            self.rwlock.release_write()

    def updatesong(self, song):
        """updates entry of given song"""
        # XXX: changes of artist, album and genre information are not handled
        self.rwlock.acquire_write()

        # update song cache if existent
        try:
            # Note that we rely on each song being uniquely specified
            # by its path and the corresponding __cmp__ method of
            # item.song!
            oldsong = self.songs[song.path]
            self.allsongs[self.allsongs.index(oldsong)] = song
        except (ValueError, AttributeError):
            pass

        self.songs[song.path] = song
        
        self.rwlock.release_write()

    def updatealbum(self, album):
        """updates entry of given album of artist"""
        # XXX: changes of other indices not handled correctly
        self.rwlock.acquire_write()
        self.albums[_albumid(album.artist, album.name)] = album
        self.rwlock.release_write()

    def updateartist(self, artist):
        """updates entry of given artist"""
        # XXX: changes of other indices not handled correctly
        self.rwlock.acquire_write()

        # update artist cache if existent
        try:
            # Note that we rely on each artist being uniquely specified
            # by its name and the corresponding __cmp__ method of
            # item.artist!
            oldartist = self.artists[artist.name]
            self.allartists[self.allartists.index(oldartist)] = artist
        except (ValueError, AttributeError):
            pass

        self.artists[artist.name] = artist
        
        self.rwlock.release_write()

    def getsong(self, path):
        """returns song entry with given path"""
        self.rwlock.acquire_read()
        song = self.songs.get(path)
        self.rwlock.release_read()
        return song

    def getalbum(self, artist, album):
        """returns song entry with given path"""
        self.rwlock.acquire_read()
        aalbum = self.albums[_albumid(artist, album)]
        self.rwlock.release_read()
        return aalbum

    def getartist(self, artist):
        """returns song entry with given path"""
        self.rwlock.acquire_read()
        aartist = self.artists.get(artist)
        self.rwlock.release_read()
        return aartist

    def getartists(self):
        """return all stored artists"""
        self.rwlock.acquire_read()
        # use cached value if existent
        if self.allartists:
            artists = self.allartists[:]
            self.rwlock.release_read()
            return artists
        else:
            debug("songdb: start rebuilding artist cache\n")
            artists = map(self.artists.get, self.artists.keys())
            debug("songdb: finished rebuilding artist cache\n")
            self.rwlock.release_read()
            newallartists = artists[:]
            self.rwlock.acquire_write()
            debug("songdb: saving artist cache\n")
            self.allartists = newallartists
            self.rwlock.release_write()
            return artists

    def getalbums(self, artist=None):
        """return albums of a given artist.

        artist has to be a string. If it is none, all stored
        albums are returned
        """
        if artist is None:
            albums = []
            albums = map(self.albums.get, self.albums.keys())
            return albums
        else:
            self.rwlock.acquire_read()
            albums = map(self.albums.get, self.artists[artist].albums)
            self.rwlock.release_read()
            return albums

    def getsongs(self, artist=None, album=None):
        """ returns song of given artist and given album
        
        artist and album are either strings or None
        If they are not given all artists and/or albums are chosen
        """
        
        if artist is None and album is None:
            # return all songs in songdb
            self.rwlock.acquire_read()
            # use cached value if existent
            if self.allsongs:
                songs = self.allsongs[:]
                self.rwlock.release_read()
                return songs
            else:
                songs = map(self.songs.get, self.songs.keys())
                self.rwlock.release_read()
                newallsongs = songs[:]
                self.rwlock.acquire_write()
                self.allsongs = newallsongs
                self.rwlock.release_write()
                return songs
        elif album is None:
            # return all songs of a given artist
            keys = self.getartist(artist).songs
            self.rwlock.acquire_read()
            songs = map(self.songs.get, keys)
            self.rwlock.release_read()
            return songs
        elif artist is None:
            # return all songs which are on albums with the given name
            return [ aalbum for aalbum in self.getalbums()
                            if aalbum.name==album ]
        else:
            # return all songs on a given album of a given artist
            keys = self.getalbum(artist, album).songs
            self.rwlock.acquire_read()
            songs = map(self.songs.get, keys)
            self.rwlock.release_read()
            return songs
        
    def getgenres(self):
        """return all stored genres"""
        self.rwlock.acquire_read()
        keys = self.genres.keys()
        self.rwlock.release_read()
        genres = map(self.genres.get, keys)
        return genres

    def getyears(self):
        """return all stored years"""
        self.rwlock.acquire_read()
        keys = self.years.keys()
        self.rwlock.release_read()
        years = map(self.years.get, keys)
        return years

    def getplaylist(self, path):
        """returns playlist entry with given path"""
        self.rwlock.acquire_read()
        playlist = self.playlists.get(path)
        self.rwlock.release_read()
        return playlist

    def getplaylists(self):
        self.rwlock.acquire_read()
        keys = self.playlists.keys()
        self.rwlock.release_read()
        playlists = map(self.getplaylist, keys)
        return playlists

    def registerdirtree(self, dir):
        for name in os.listdir(dir):
            path = os.path.join(dir, name)
            if os.access(path, os.R_OK):
                if os.path.isdir(path):
                    try: self.registerdirtree(path)
                    except (IOError, OSError): pass
                elif name.endswith(".mp3"):
                    try: self.queryregistersong(path)
                    except (IOError, OSError): pass
                elif name.endswith(".ogg") and item.oggsupport:
                    try: self.queryregistersong(path)
                    except (IOError, OSError): pass
                elif name.endswith(".m3u"):
                    try: self.registerplaylist(path)
                    except (IOError, OSError): pass

    # event handler

    def songupdate(self, event):
        self.updatesong(event.song)

#
# thread for automatic registering of songs in database
#

class songautoregisterer(threading.Thread):
    
    def __init__(self, db, basedir):
        self.db = db
        self.basedir = basedir
        self.stop = 0
        threading.Thread.__init__(self)

    def registerdirtree(self, dir):
        if self.stop:
            return
        for name in os.listdir(dir):
            path = os.path.join(dir, name)
            if os.access(path, os.R_OK):
                if os.path.isdir(path):
                    try: self.registerdirtree(path)
                    except (IOError, OSError): pass
                elif name.endswith(".mp3"):
                    try: self.db.queryregistersong(path)
                    except (IOError, OSError): pass
                elif name.endswith(".ogg") and item.oggsupport:
                    try: self.db.queryregistersong(path)
                    except (IOError, OSError): pass
                elif name.endswith(".m3u"):
                    try: self.db.registerplaylist(path)
                    except (IOError, OSError): pass

    def run(self):
        debug("songautoregisterer: registering songs\n")
        self.registerdirtree(self.basedir)
        if self.stop:
            return

        # fill caches
        debug("songautoregisterer: filling song cache\n")
        self.db.getsongs()
        
        debug("songautoregisterer: filling artist cache\n")
        self.db.getartists()
        
        debug("songautoregisterer: exit\n")
